/* tabulate.cc - A program for visualising data from log files containing
 *               benchmark timings generated by vnus executables.
 *               Author: Leo Breebaart, October 2001
*/

#include <map>
#include <vector>
#include <string>
#include <cstdlib>
#include <cstring>
#include <getopt.h>


    // Types:

enum Outputformat { OUT_STANDARD, OUT_LATEX, OUT_GNUPLOT };

    // This struct accumulates timing info for runs using a given number
    // of processors.
struct runinfo
{
    int count;      // Total number of collected runs
    double minimum; // Minimum thread execution time
    double maximum; // Maximum thread execution time
    double average; // Average thread execution time
};


    // Globals:

static enum Outputformat outputformat = OUT_STANDARD;

static char *progname;

    // Map key will be the number of processors, -1 indicates sequential run.
static std::map <int, struct runinfo> runs;
static const int SEQUENTIAL = -1;

static std::vector <FILE *> logfiles;

static FILE *outfile = stdout;


    // Functions:

static void handle_options (const int argc, char * const argv[]);
static void collect_info ();
static void print_table ();
static void cleanup ();
static void print_usage ();


static void print_usage ()
{
    printf("usage: tabulate [<flags>] <logfiles>\n");
    printf(" -f<type>  Set output format to type <type>\n");
    printf(" -o<file>  Redirect output to file <file>\n");
    printf("\n");
    printf("output format types:\n");
    printf(" 0, ascii   - ASCII tabular format (default)\n");
    printf(" 1, latex   - LaTeX table format\n");
    printf(" 2, gnuplot - Gnuplot source format\n");
}


int main (const int argc, char * const argv[])
{
    handle_options(argc, argv);

    collect_info();

    print_table();

    cleanup();

    exit(EXIT_SUCCESS);
}


static void handle_options (const int argc, char * const argv[])
{
    progname = argv[0];

    int c;

    while (true)
    {
        if ((c = getopt(argc, argv, "f:o:h")) == -1)
            break;

        switch (c)
        {
                // Set output format.
            case 'f':

                if (strstr(optarg, "ascii") == &optarg[0])
                    outputformat = OUT_STANDARD;
                else if (strstr(optarg, "latex") == &optarg[0])
                    outputformat = OUT_LATEX;
                else if (strstr(optarg, "gnuplot") == &optarg[0])
                    outputformat = OUT_GNUPLOT;
                else
                    outputformat = (enum Outputformat)(atoi(optarg));

                switch (outputformat)
                {
                    case OUT_STANDARD:
                    case OUT_LATEX:
                    case OUT_GNUPLOT:
                        break;

                    default:
                        fprintf(stderr, "Error: argument to -f must be 0-2, but was: %d\n", outputformat);
                        exit(0);
                        break;
                }
                break;

                    // Redirect output to file.
            case 'o':

                if ((outfile = fopen(optarg, "w")) == NULL)
                {
                    fprintf(stderr, "%s: error opening output file '%s' for writing\n", progname, optarg);
                    perror(optarg);
                    exit(2);
                }
                break;

            case 'h':
                print_usage();
                exit(0);
                break;

                    // getopt error conditions
            case ':':
            case '?':
                print_usage();
                exit(2);
                break;

            default:
                fprintf(stderr, "%s: getopt returned unknown character code 0%o\n", progname, c);
                exit(1);
                break;
        }
    }

        // Everything left over should be a logfile for input.
    if (optind >= argc)
    {
        fprintf(stderr, "%s: no log files specified\n", progname);
        print_usage();
        exit(1);
    }
    else
    {
        while (optind < argc)
        {
            FILE *logfile;
            char *logfilename = argv[optind++];

            if ((logfile = fopen(logfilename, "r")) == NULL)
            {
                fprintf(stderr, "%s: error opening logfile '%s' for reading\n", progname, logfilename);
                perror(logfilename);
                exit(2);
            }

            logfiles.push_back(logfile);
        }
    }
}


#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

static void collect_info ()
{
    std::vector <FILE *>::iterator f;
    for (f = logfiles.begin(); f != logfiles.end(); f++)
    {
	char buf[256];

	while (fgets(buf, 256, (*f)) != NULL)
	{
                // Empty lines and comments are skipped
            if (strchr(buf, '\n') == &buf[0] || strchr(buf, '#') == &buf[0])
                continue;

	    int procnr;
	    double minimum, maximum, average;

                // Important note: run file lines are in 'avg min max'
                // format, but tabulate.cc itself was changed to output
                // 'min avg max' for all formats.
	    sscanf(buf, "%d  %lf  %lf  %lf", &procnr, &average, &minimum, &maximum);

            struct runinfo &r = runs[procnr];

	    r.count++;

            if (r.count == 1)
            {
                r.minimum = minimum;
                r.maximum = maximum;
            }
            else
            {
                r.minimum = min(r.minimum, minimum);
                r.maximum = max(r.maximum, maximum);
            }

	    r.average += average;
	}
    }

    std::map<int, struct runinfo>::iterator it;
    for (it = runs.begin(); it != runs.end(); it++)
    {
	struct runinfo &r = (*it).second;

	r.average = r.average / r.count;
    }
}


static void print_table ()
{
        // Print table headers.
    switch (outputformat)
    {
        case OUT_STANDARD:

            fprintf(outfile, "Benchmark table:\n");
            break;

        case OUT_LATEX:

            fprintf(outfile, "\n");
            fprintf(outfile, "{\\scriptsize");
            fprintf(outfile, "\\begin{tabular}{|r|rrr|}\n");
            fprintf(outfile, "\\hline\n");
            fprintf(outfile, "\\#p & min & avg & max \\\\\n");
            fprintf(outfile, "\\hline\n");
            fprintf(outfile, "\\hline\n");
            break;

        case OUT_GNUPLOT:
            fprintf(outfile, "# Gnuplot data:\n");
            break;

        default:
            break;
    }

        // Print table rows for parallel runs (p != SEQUENTIAL).
    std::map<int, struct runinfo>::iterator it;
    for (it = runs.begin(); it != runs.end(); it++)
    {
	int nproc = (*it).first;
	struct runinfo r = (*it).second;

        if (nproc == SEQUENTIAL)
            continue;

        switch (outputformat)
        {
            case OUT_STANDARD:

                fprintf(outfile, "#proc = %2d, #runs = %2d, min = %4.1f, avg = %4.1f, max = %4.1f\n",
                        nproc, r.count, r.minimum, r.average, r.maximum);
                break;

            case OUT_LATEX:

                fprintf(outfile, "%2d & %.1f & %.1f & \\textbf{%.1f} \\\\\n",
                        nproc, r.minimum, r.average, r.maximum);
                fprintf(outfile, "\\hline\n");
                break;

            case OUT_GNUPLOT:
                fprintf(outfile, "%2d %.1f %.1f %.1f\n", nproc, r.minimum, r.average, r.maximum);
                break;

            default:
                break;
        }
    }

        // Was there a sequential run listed as well?

    if (runs.find(SEQUENTIAL) != runs.end())
    {
	struct runinfo r = runs[SEQUENTIAL];

        switch (outputformat)
        {
            case OUT_STANDARD:

                fprintf(outfile, "#proc = seq, #runs = %2d, min = %4.1f, avg = %4.1f, max = %4.1f\n",
                        r.count, r.minimum, r.average, r.maximum);
                break;

            case OUT_LATEX:

                fprintf(outfile, "\\hline\n");
                fprintf(outfile, "seq & %.1f & %.1f & \\textbf{%.1f} \\\\\n",
                        r.minimum, r.average, r.maximum);
                fprintf(outfile, "\\hline\n");
                break;

            case OUT_GNUPLOT:
                break;

            default:
                break;
        }
    }


        // Print table end bits.
    switch (outputformat)
    {
        case OUT_STANDARD:
            break;

        case OUT_LATEX:

            fprintf(outfile, "\\end{tabular}\n");
            fprintf(outfile, "}");
            fprintf(outfile, "\n");
            break;

        case OUT_GNUPLOT:
            break;

        default:
            break;
    }
}


static void cleanup ()
{
    std::vector <FILE *>::iterator f;
    for (f = logfiles.begin(); f != logfiles.end(); f++)
    {
	if (fclose(*f) != 0)
	{
	    fprintf(stderr, "%s: error closing logfile\n", progname);
	}
    }

    if (fclose(outfile) != 0)
    {
	fprintf(stderr, "%s: error closing output file\n", progname);
	perror("Error: ");
    }
}
